"""
Rest API paste factory and related.
"""
import os
import urllib

import zope.configuration.config
import zope.configuration.xmlconfig
from lazr.restful.wsgi import BaseWSGIWebServiceConfiguration
from lazr.restful.publisher import WebServicePublicationMixin
from lazr.restful.publisher import WebServiceRequestTraversal
from zope.app.wsgi import WSGIPublisherApplication
from zope.app.publication.zopepublication import ZopePublication
from zope.app.publication.httpfactory import HTTPPublicationRequestFactory
from zope.app.publication.httpfactory import chooseClasses as zopeAppChooseClasses
from zope.component import provideUtility
from zope.site import threadSiteSubscriber
from zope.publisher.publish import mapply
from zope.publisher.browser import BrowserRequest
from zope.publisher.http import HTTPRequest
from zope.traversing.interfaces import BeforeTraverseEvent
from ZODB.interfaces import IDatabase

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.main import SchoolToolMachinery, setLanguage


class WebServiceConfiguration(BaseWSGIWebServiceConfiguration):
    path_override = None
    code_revision = '1'
    active_versions = ['1.0']
    use_https = False
    last_version_with_mutator_named_operations = None
    view_permission = 'schooltool.view'


class ImmediateObjectCaller(object):

    def callObject(self, request, ob):
        return mapply(ob, request.getPositionalArguments(), request)


class SchoolToolServicePublicationMixin(WebServicePublicationMixin, ImmediateObjectCaller):

    application = None

    def getSchoolApplication(self, request):
        return super(SchoolToolServicePublicationMixin, self).getApplication(request)

    def getApplication(self, request):
        return self.application


class SchoolToolServiceRequestTraversal(WebServiceRequestTraversal):

    _web_service_names = None

    def getWebServiceURL(self, depth=0, path_only=False):
        app_names = self._app_names
        if self._web_service_names is not None:
            app_names = list(self._web_service_names + app_names)
        if depth:
            names = self._traversed_names
            if depth > len(names):
                raise IndexError(depth)
            names = app_names + names[:depth]
        else:
            names = app_names

        # See: http://www.ietf.org/rfc/rfc2718.txt, Section 2.2.5
        names = [urllib.quote(name.encode("utf-8"), safe='/+@')
                 for name in names]
        if path_only:
            return names and ('/' + '/'.join(names)) or '/'
        return "/".join([self._app_server] + names)


REQUEST_BASES = {
    HTTPRequest: BrowserRequest,
    }

PUBLICATION_BASES = {
    }

class RestfulPublicationRequestFactory(HTTPPublicationRequestFactory):

    last_req_cls = None
    last_restful_req = None

    def __call__(self, input_stream, env):
        method = env.get('REQUEST_METHOD', 'GET').upper()
        req_cls, pub_cls = zopeAppChooseClasses(method, env)
        pub_cls = PUBLICATION_BASES.get(pub_cls, pub_cls)
        publication = self._publication_cache.get(pub_cls)
        if publication is None:
            restful_cls = type(
                'Restful'+pub_cls.__name__,
                (SchoolToolServicePublicationMixin, pub_cls, ),
                {})
            publication = restful_cls(self._db)
            self._publication_cache[pub_cls] = publication

        req_cls = REQUEST_BASES.get(req_cls, req_cls)
        if self.last_req_cls == req_cls:
            restful_req_cls = self.last_restful_req
        else:
            restful_req_cls = type(
                'Restful'+req_cls.__name__,
                (SchoolToolServiceRequestTraversal, req_cls, ),
                {})
            self.last_req_cls = req_cls
            self.last_restful_req = restful_req_cls

        request = restful_req_cls(input_stream, env)
        request.setPublication(publication)
        request._web_service_names = request._app_names
        return request


class RestAPIServer(SchoolToolMachinery, WSGIPublisherApplication):

    config = None
    config_handler = None
    db = None
    standalone = False

    def __init__(self, config_filename, factory=RestfulPublicationRequestFactory):
        self.setup(config_filename)
        super(RestAPIServer, self).__init__(self.db, factory)

    def setup(self, config_filename):
        options = self.loadOptions(config_filename)
        db = self.openDB(options)
        if self.standalone:
            self.configureComponents(options)
            setLanguage(options.config.lang)
        return db

    def loadOptions(self, config_filename):
        options = self.Options()
        options.config_file = config_filename
        options.config, options.config_handler = self.readConfig(options.config_file)
        options.zcml_context = zope.configuration.config.ConfigurationMachine()
        if options.config.database.config.storage is None:
            raise Exception('No ZODB storage configured')
        return options

    def openDB(self, options):
        if self.db is not None:
            return self.db
        db_configuration = options.config.database
        db = db_configuration.open()
        connection = db.open()
        root = connection.root()
        app = root.get(ZopePublication.root_name)
        if app is None or not ISchoolToolApplication.providedBy(app):
            connection.close()
            return None
        connection.close()
        provideUtility(db, IDatabase)
        self.db = db
        return self.db

    def configureComponents(self, options):
        if options.config.devmode:
            options.zcml_context.provideFeature('devmode')
        zope.configuration.xmlconfig.registerCommonDirectives(options.zcml_context)
        # reuse site definition for now
        options.zcml_context = zope.configuration.xmlconfig.file(
            options.config.site_definition, context=options.zcml_context)


def traverseSchoolToolAPISubscriber(ob, event):
    app = event.request.publication.getSchoolApplication(event.request)
    event = BeforeTraverseEvent(app, event.request)
    threadSiteSubscriber(app, event)
    event.request.publication._maybePlacefullyAuthenticate(event.request, app)


rest_api_app = None

def rest_api_factory(global_conf, config_file):
    global rest_api_app
    if rest_api_app:
        return
    rest_api_app = RestAPIServer(os.path.join(global_conf['here'], config_file))
    return rest_api_app
